/*
 * Filter
 * Copyright (C) 2011 by Apaq
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package dk.apaq.filter.mongodb;

import dk.apaq.filter.Filter;
import dk.apaq.filter.core.AndFilter;
import dk.apaq.filter.core.CompareFilter;
import dk.apaq.filter.core.NotFilter;
import dk.apaq.filter.core.OrFilter;
import dk.apaq.filter.limit.Limit;
import dk.apaq.filter.mongodb.interpreters.CompareFilterInterpreter;
import dk.apaq.filter.mongodb.interpreters.Interpreter;
import dk.apaq.filter.sort.SortDirection;
import dk.apaq.filter.sort.Sorter;
import dk.apaq.filter.sort.SorterEntry;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Order;
import org.springframework.data.mongodb.core.query.Query;

/**
 *  Translates Filter, Sorters and Limits to JPA Query.
 */
public class FilterTranslatorForMongoDb {

    private static final Map<Class<? extends Filter>, Interpreter<? extends Filter>> interpreterMap = new HashMap<Class<? extends Filter>, Interpreter<? extends Filter>>();

    static {
        interpreterMap.put(CompareFilter.class, new CompareFilterInterpreter());
    }
    
    public static Query translate(Class entityClass, Filter filter, Sorter sorter, Limit limit) {
        return translate(null, entityClass, filter, sorter, limit);
    }
    
    public static Query translate(String[] fields, Class entityClass, Filter filter, Sorter sorter, Limit limit) {
    
        Query q = new Query();
    
        // Handle fields
        if(fields != null) {
            for(String field : fields) {
                q.fields().include(field);
            }
        }
                
        //Handle filter
        Criteria c = new Criteria();
        buildCriteria(c, filter);
        q.addCriteria(c);
        
        //Handle sorting
        if(sorter != null && ! sorter.getSorterEntries().isEmpty()) {
            if(sorter.getSorterEntries().size()>1) {
                throw new IllegalArgumentException("One a single sorter element is supported by this translator.");
            }
            
            SorterEntry entry = sorter.getSorterEntries().get(0);
            q.sort().on(entry.getPropertyId(), entry.getDirection() == SortDirection.Ascending ? Order.ASCENDING : Order.DESCENDING);
        }
        
        //Handle limits
        if(limit != null) {
            q.skip(limit.getOffset());
            q.limit(limit.getCount());
        }
        
        return q;

    }
    
    private static boolean buildCriteria(Criteria clause, Filter filter) {

        if (filter == null) {
            return false;
        }

        //First try an interpreter
        Interpreter interpreter = interpreterMap.get(filter.getClass());
        if (interpreter != null) {
            return interpreter.interpret(clause, filter);
        }


        //..else try our special handlers for filtercontainers.
        if (filter instanceof AndFilter) {
            AndFilter af = (AndFilter) filter;
            List<Criteria> criterias = new ArrayList<Criteria>();
            boolean hasAddedFilter = false;
            
            for (Filter currentFilter : af.getFilters()) {
                Criteria c = new Criteria();
                boolean statementAppended = buildCriteria(c, currentFilter);

                if (statementAppended) {
                    criterias.add(c);
                    hasAddedFilter = true;
                }
            }

            clause.andOperator(criterias.toArray(new Criteria[0]));
            return hasAddedFilter;
        }

        if (filter instanceof OrFilter) {
            OrFilter of = (OrFilter) filter;
            List<Criteria> criterias = new ArrayList<Criteria>();
            boolean hasAddedFilter = false;

            for (Filter currentFilter : of.getFilters()) {
                Criteria c = new Criteria();
                boolean statementAppended = buildCriteria(c, currentFilter);

                if (statementAppended) {
                    criterias.add(c);
                    hasAddedFilter = true;
                }
            }

            clause.orOperator(criterias.toArray(new Criteria[0]));
            return hasAddedFilter;
        }

        if (filter instanceof NotFilter) {
            NotFilter af = (NotFilter) filter;

            clause.not();
            return buildCriteria(clause, af.getFilterElement());
        }

        return false;
    }
}
